package bk;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collections;

import javax.swing.JTree;

import animation.AlgorithmStage;
import animation.BackwardAction;
import animation.PseudoCodeNode;
import animation.PseudoCodeNode.CodeAction;
import bk.ExtremalLayoutCalc.LayoutType;
import graph.LayeredGraphNode;
import lib.TextLayoutHelper;

/**
 * The stage of the combination of the four extremal layouts.
 * @author kolja
 *
 */
public class Combine implements AlgorithmStage {

    LayeredGraphNode graph;

    private enum State
    {
        ALIGN,
        SET_COORDINATES
    }

    private State state;
    private int btlOffset;
    private int btrOffset;
    private int tblOffset;
    private int tbrOffset;
    private int vIndex;
    private ArrayList< BackwardAction > actions;
    private PseudoCodeNode alignNode;
    private PseudoCodeNode setNode;
    private PseudoCodeNode loopNode;
    private boolean inside;
    private boolean insideSubgraph;
    private boolean breakPoint;
    private boolean skipBackwards;
    private ArrayList< PseudoCodeNode > subgraphNodes;
    private ArrayList< Combine > subgraphAlgs;

    public Combine( LayeredGraphNode graph )
    {
        this.graph = graph;
        state = State.ALIGN;
        vIndex = 0;
        actions = new ArrayList<>();
        inside = false;
        insideSubgraph = false;
        skipBackwards = false;
        subgraphNodes = new ArrayList<>();
        subgraphAlgs = new ArrayList<>();
        for( @SuppressWarnings("unused") LayeredGraphNode n : graph.getContainedNodes() )
        {
            subgraphAlgs.add( null );
            subgraphNodes.add( null );
        }
    }

    @Override
    public StageStatus forwardStep() {
        breakPoint = false;
        if( state == State.ALIGN )
        {
            inside = false;
            boolean sel = alignNode.isSelected();
            CodeAction action = alignNode.setSelected( true );
            if( !sel )
                breakPoint |= action == CodeAction.STOP;
            int tblw = (int)calcMaxX( LayoutType.TOP_BOTTOM_LEFT );
            int tbrw = (int)calcMaxX( LayoutType.TOP_BOTTOM_RIGHT );
            int btlw = (int)calcMaxX( LayoutType.BOTTOM_TOP_LEFT );
            int btrw = (int)calcMaxX( LayoutType.BOTTOM_TOP_RIGHT );
            LayoutType minLayout = LayoutType.TOP_BOTTOM_LEFT;
            int minWidth = tblw;
            if( tbrw < minWidth )
            {
                minWidth = tbrw;
                minLayout = LayoutType.TOP_BOTTOM_RIGHT;
            }
            if( btlw < minWidth )
            {
                minWidth = btlw;
                minLayout = LayoutType.BOTTOM_TOP_LEFT;
            }
            if( btrw < minWidth )
            {
                minWidth = btrw;
                minLayout = LayoutType.BOTTOM_TOP_RIGHT;
            }
            int minX = calcMinX( minLayout );
            btlOffset = minX - calcMinX( LayoutType.BOTTOM_TOP_LEFT );
            tblOffset = minX - calcMinX( LayoutType.TOP_BOTTOM_LEFT );
            btrOffset = minWidth - btrw;
            tbrOffset = minWidth - tbrw;
            Color oldColor = graph.getColor( LayoutType.COMBINED );
            if( oldColor == null )
                graph.setColor( Color.BLACK, null );
            actions.add( 0, () -> {
                inside = false;
                loopNode.setSelected( false );
                setNode.setSelected( false );
                breakPoint = false;
                boolean selected = alignNode.isSelected();
                CodeAction ac = alignNode.setSelected( true );
                if( !selected )
                    breakPoint |= ac == CodeAction.STOP;
                skipBackwards = ac == CodeAction.SKIP;
                state = State.ALIGN;
                if( oldColor == null )
                    graph.setColor( null, null );
            });
            state = State.SET_COORDINATES;
            alignNode.setSelected( false );
            sel = setNode.isSelected();
            CodeAction ac = setNode.setSelected( true );
            boolean skip = ac == CodeAction.SKIP;
            if( !sel )
                breakPoint |= ac == CodeAction.STOP;
            ac = loopNode.setSelected( true );
            breakPoint |= ac == CodeAction.STOP;
            skip |= ac == CodeAction.SKIP;
            if( (action == CodeAction.SKIP || skip) && !breakPoint )
                return forwardStep();
        }
        else
        {
            CodeAction action = loopNode.setSelected( true );
            breakPoint = action == CodeAction.STOP;
            if( vIndex >= graph.getContainedNodes().size() )
            {
                inside = false;
                setNode.setSelected( false );
                loopNode.setSelected( false );
                return StageStatus.FINISHED;
            }
            else
                inside = true;
            LayeredGraphNode current = graph.getContainedNodes().get( vIndex );
            current.setSelected( null );
            if( current.getContainedNodes().size() > 0 )
            {
                insideSubgraph = true;
                boolean breakpoint = false;
                boolean sel = subgraphNodes.get( vIndex ).isSelected();
                CodeAction ac = subgraphNodes.get( vIndex ).setSelected( true );
                if( !sel )
                    breakpoint |= ac == CodeAction.STOP;
                do {
                    switch( subgraphAlgs.get( vIndex ).forwardStep() )
                    {
                    case BREAKPOINT:
                        return StageStatus.BREAKPOINT;
                    case UNFINISHED:
                        break;
                    case FINISHED:
                        insideSubgraph = false;
                        subgraphNodes.get( vIndex ).setSelected( false );
                        ac = null;
                        break;
                    }
                } while( ac == CodeAction.SKIP && !breakpoint );
            }
            ArrayList< Integer > positions = new ArrayList<>();
            positions.add( (Integer)(int)current.getX( LayoutType.TOP_BOTTOM_LEFT ) + tblOffset );
            positions.add( (Integer)(int)current.getX( LayoutType.TOP_BOTTOM_RIGHT ) + tbrOffset );
            positions.add( (Integer)(int)current.getX( LayoutType.BOTTOM_TOP_LEFT ) + btlOffset );
            positions.add( (Integer)(int)current.getX( LayoutType.BOTTOM_TOP_RIGHT ) + btrOffset );
            Collections.sort( positions );
            int oldX = (int)current.getX( LayoutType.COMBINED );
            current.setX( (positions.get( 1 ) + positions.get( 2 )) / 2, true, LayoutType.COMBINED );
            actions.add( 0, () -> {
                inside = true;
                breakPoint = false;
                boolean selected = setNode.isSelected();
                CodeAction ac = setNode.setSelected( true );
                if( !selected )
                    breakPoint |= ac == CodeAction.STOP;
                skipBackwards = ac == CodeAction.SKIP;
                ac = loopNode.setSelected( true );
                breakPoint |= ac == CodeAction.STOP;
                skipBackwards |= ac == CodeAction.SKIP;
                vIndex--;
                current.setX( oldX, true, LayoutType.COMBINED );
                current.setSelected( null );
            });
            vIndex++;
            if( action == CodeAction.SKIP && !breakPoint )
                return forwardStep();
        }
        if( breakPoint )
            return StageStatus.BREAKPOINT;
        return StageStatus.UNFINISHED;
    }

    private int calcMinX(  LayoutType layout )
    {
        int minX = 0;
        if( graph.getContainedNodes().size() > 0 )
            minX = (int)graph.getContainedNodes().get( 0 ).getX( layout );
        for( LayeredGraphNode n : graph.getContainedNodes() )
            minX = Math.min( minX, (int)n.getX( layout ) );
        return minX;
    }

    private int calcMaxX(  LayoutType layout )
    {
        int maxX = 0;
        if( graph.getContainedNodes().size() > 0 )
            maxX = (int)graph.getContainedNodes().get( 0 ).getX( layout );
        for( LayeredGraphNode n : graph.getContainedNodes() )
            maxX = Math.max( maxX, (int)n.getX( layout ) );
        return maxX;
    }

    @Override
    public StageStatus backwardStep() {
        if( vIndex < subgraphAlgs.size() && subgraphAlgs.get( vIndex ) != null )
        {
            insideSubgraph = true;
            boolean breakpoint = false;
            boolean selected = subgraphNodes.get( vIndex ).isSelected();
            CodeAction action = subgraphNodes.get( vIndex ).setSelected( true );
            if( !selected )
                breakpoint |= action == CodeAction.STOP;
            do {
                switch( subgraphAlgs.get( vIndex ).backwardStep() )
                {
                case BREAKPOINT:
                    return StageStatus.BREAKPOINT;
                case UNFINISHED:
                    LayeredGraphNode current = graph.getContainedNodes().get( vIndex );
                    current.setSelected( null );
                    break;
                case FINISHED:
                    insideSubgraph = false;
                    subgraphNodes.get( vIndex ).setSelected( false );
                    action = null;
                    break;
                }
            } while( !breakpoint && action == CodeAction.SKIP );
        }
        if( actions.size() == 0 )
        {
            inside = false;
            alignNode.setSelected( false );
            return StageStatus.FINISHED;
        }
        actions.get( 0 ).reverse();
        actions.remove( 0 );
        if( breakPoint )
            return StageStatus.BREAKPOINT;
        if( skipBackwards )
            return backwardStep();
        return StageStatus.UNFINISHED;
    }

    @Override
    public PseudoCodeNode createPseudocodeTree( JTree tree ) {
        PseudoCodeNode root = new PseudoCodeNode( TextLayoutHelper.setupPseudoCodeStage("Balancing"), tree );
        alignNode = new PseudoCodeNode( TextLayoutHelper.setupPseudoCodeStage("Align Layouts"), tree );
        setNode = new PseudoCodeNode( TextLayoutHelper.setupPseudoCodeStage("Align to assignment of smallest width"), tree );
        loopNode = new PseudoCodeNode( "Set coordinates to average median of aligned candidates", tree );
        for( int i = 0; i < graph.getContainedNodes().size(); i++ )
        {
            LayeredGraphNode current = graph.getContainedNodes().get( i );
            if( current.getContainedNodes().size() > 0 )
            {
                Combine extcalc = new Combine( current );
                PseudoCodeNode subNode = extcalc.createPseudocodeTree( loopNode.getTree() );
                loopNode.add( subNode );
                subgraphAlgs.set( i, extcalc );
                subgraphNodes.set( i, subNode );
            }
        }
        setNode.add( loopNode );
        root.add( alignNode );
        root.add( setNode );
        return root;
    }

    @Override
    public StageStatus forwardStepOver() {
        if( !inside )
        {
            State oldState = state;
            StageStatus stage = StageStatus.UNFINISHED;
            while( state == oldState && stage == StageStatus.UNFINISHED )
                stage = forwardStep();
            return stage;
        }
        else
        {
            if( !insideSubgraph )
                return forwardStep();
            else
            {
                boolean breakpoint = false;
                boolean selected = subgraphNodes.get( vIndex ).isSelected();
                CodeAction action = subgraphNodes.get( vIndex ).setSelected( true );
                if( !selected )
                    breakpoint |= action == CodeAction.STOP;
                do {
                    switch( subgraphAlgs.get( vIndex ).forwardStepOver() )
                    {
                    case BREAKPOINT:
                        return StageStatus.BREAKPOINT;
                    case UNFINISHED:
                        break;
                    case FINISHED:
                        insideSubgraph = false;
                        subgraphNodes.get( vIndex ).setSelected( false );
                        action = null;
                        break;
                    }
                } while( !breakpoint && action == CodeAction.SKIP );
                return StageStatus.UNFINISHED;
            }
        }
    }

    @Override
    public StageStatus forwardStepOut() {
        if( !inside )
        {
            StageStatus status = StageStatus.UNFINISHED;
            while( status == StageStatus.UNFINISHED )
                status = forwardStep();
            return status;
        }
        else
        {
            if( !insideSubgraph )
            {
                State oldState = state;
                StageStatus stage = StageStatus.UNFINISHED;
                while( state == oldState && stage == StageStatus.UNFINISHED )
                    stage = forwardStep();
                return stage;
            }
            else
            {
                boolean breakpoint = false;
                boolean selected = subgraphNodes.get( vIndex ).isSelected();
                CodeAction action = subgraphNodes.get( vIndex ).setSelected( true );
                if( !selected )
                    breakpoint |= action == CodeAction.STOP;
                do {
                    switch( subgraphAlgs.get( vIndex ).forwardStepOut() )
                    {
                    case BREAKPOINT:
                        return StageStatus.BREAKPOINT;
                    case UNFINISHED:
                        break;
                    case FINISHED:
                        insideSubgraph = false;
                        subgraphNodes.get( vIndex ).setSelected( false );
                        action = null;
                        break;
                    }
                } while( !breakpoint && action == CodeAction.SKIP );
                return StageStatus.UNFINISHED;
            }
        }
    }

    @Override
    public StageStatus backwardStepOver() {
        if( !inside )
        {
            State oldState = state;
            StageStatus stage = StageStatus.UNFINISHED;
            while( state == oldState && stage == StageStatus.UNFINISHED )
                stage = backwardStep();
            return stage;
        }
        else
        {
            if( !insideSubgraph )
                return backwardStep();
            else
            {
                boolean breakpoint = false;
                boolean selected = subgraphNodes.get( vIndex ).isSelected();
                CodeAction action = subgraphNodes.get( vIndex ).setSelected( true );
                if( !selected )
                    breakpoint |= action == CodeAction.STOP;
                do {
                    switch( subgraphAlgs.get( vIndex ).backwardStepOver() )
                    {
                    case BREAKPOINT:
                        return StageStatus.BREAKPOINT;
                    case UNFINISHED:
                        break;
                    case FINISHED:
                        insideSubgraph = false;
                        subgraphNodes.get( vIndex ).setSelected( false );
                        action = null;
                        break;
                    }
                } while( !breakpoint && action == CodeAction.SKIP );
                return StageStatus.UNFINISHED;
            }
        }
    }

    @Override
    public StageStatus backwardStepOut() {
        if( !inside )
        {
            StageStatus status = StageStatus.UNFINISHED;
            while( status == StageStatus.UNFINISHED )
                status = backwardStep();
            return status;
        }
        else
        {
            if( !insideSubgraph )
            {
                State oldState = state;
                StageStatus stage = StageStatus.UNFINISHED;
                while( state == oldState && stage == StageStatus.UNFINISHED )
                    stage = backwardStep();
                return stage;
            }
            else
            {
                boolean breakpoint = false;
                boolean selected = subgraphNodes.get( vIndex ).isSelected();
                CodeAction action = subgraphNodes.get( vIndex ).setSelected( true );
                if( !selected )
                    breakpoint |= action == CodeAction.STOP;
                do {
                    switch( subgraphAlgs.get( vIndex ).backwardStepOut() )
                    {
                    case BREAKPOINT:
                        return StageStatus.BREAKPOINT;
                    case UNFINISHED:
                        break;
                    case FINISHED:
                        insideSubgraph = false;
                        subgraphNodes.get( vIndex ).setSelected( false );
                        action = null;
                        break;
                    }
                } while( !breakpoint && action == CodeAction.SKIP );
                return StageStatus.UNFINISHED;
            }
        }
    }
    
    @Override
    public String getDebugString()
    {
        String info = "| Node |  x  | x LU | x RU | x LL | x RL |\n";
        info +=       "|------|-----|------|------|------|------|\n";
        for( LayeredGraphNode n : graph.getContainedNodes() )
        {
            info += "|" + TextLayoutHelper.strToLen( n.getName(), 6 ) + 
                    "|" + TextLayoutHelper.strToLen( n.getX( LayoutType.COMBINED ) + "", 5 ) + 
                    "|" + TextLayoutHelper.strToLen( ( n.getX( LayoutType.TOP_BOTTOM_LEFT ) + tblOffset ) + "", 6 ) + 
                    "|" + TextLayoutHelper.strToLen( ( n.getX( LayoutType.TOP_BOTTOM_RIGHT ) + tbrOffset ) + "", 6 ) + 
                    "|" + TextLayoutHelper.strToLen( ( n.getX( LayoutType.BOTTOM_TOP_LEFT ) + btlOffset ) + "", 6 ) + 
                    "|" + TextLayoutHelper.strToLen( ( n.getX( LayoutType.BOTTOM_TOP_RIGHT ) + btrOffset ) + "", 6 ) + "|\n";
        }
        if( insideSubgraph && vIndex < graph.getContainedNodes().size() )
        {
            info += "Subgraph of " + graph.getContainedNodes().get( vIndex ).getName() + ":\n";
            String tmp = subgraphAlgs.get( vIndex ).getDebugString();
            info += tmp;
            return info;
        }
        return info;
    }
}
