package bk;

import java.lang.reflect.InvocationTargetException;

import javax.swing.JTree;

import animation.AlgorithmStage;
import animation.PseudoCodeNode;
import animation.PseudoCodeNode.CodeAction;
import graph.LayeredGraphNode;
import lib.TextLayoutHelper;

/**
 * The stage where the for extremal layouts are computed.
 * @author kolja
 *
 */
public class ExtremalLayoutCalc implements AlgorithmStage {
    /**
     * There are four types of layouts, based on iteration order and median alignment.
     * Additionally there is one layout for the combined coordinates.
     * @author kolja
     *
     */
    public enum LayoutType{
        TOP_BOTTOM_LEFT,
        TOP_BOTTOM_RIGHT,
        BOTTOM_TOP_LEFT,
        BOTTOM_TOP_RIGHT,
        COMBINED
    }

    private enum LayoutState
    {
        BLOCK_CALCULATION,
        COMPACTION
    }

    private PseudoCodeNode pseudoCode;
    private BlockCalc bc;
    private Compaction cp;
    private LayoutState status;
    private PseudoCodeNode bcNode;
    private PseudoCodeNode cpNode;
    private LayoutType type;
    private boolean inside;
    private LayeredGraphNode graph;

    public ExtremalLayoutCalc( LayoutType typ, LayeredGraphNode graph )
    {
        this.graph = graph;
        type = typ;
        status = LayoutState.BLOCK_CALCULATION;
        bc = new BlockCalc( graph, typ );
        cp = new Compaction( graph, typ );
        inside = false;
    }

    @Override
    public StageStatus forwardStep() {
        return forward( "forwardStep" );
    }

    @Override
    public StageStatus backwardStep() {
        return backward( "backwardStep" );
    }

    @Override
    public PseudoCodeNode createPseudocodeTree( JTree tree ) {
        if( type == LayoutType.TOP_BOTTOM_LEFT )
            pseudoCode = new PseudoCodeNode( TextLayoutHelper.setupPseudoCodeStage("Extremal layout: leftmost upper"), tree );
        if( type == LayoutType.TOP_BOTTOM_RIGHT )
            pseudoCode = new PseudoCodeNode( TextLayoutHelper.setupPseudoCodeStage("Extremal layout: rightmost upper"), tree );
        if( type == LayoutType.BOTTOM_TOP_LEFT )
            pseudoCode = new PseudoCodeNode( TextLayoutHelper.setupPseudoCodeStage("Extremal layout: leftmost lower"), tree );;
        if( type == LayoutType.BOTTOM_TOP_RIGHT )
            pseudoCode = new PseudoCodeNode( TextLayoutHelper.setupPseudoCodeStage("Extremal layout: rightmost lower"), tree );
        bcNode = bc.createPseudocodeTree( tree );
        cpNode = cp.createPseudocodeTree( tree );
        pseudoCode.add( bcNode );
        pseudoCode.add( cpNode );
        return pseudoCode;
    }

    @Override
    public StageStatus forwardStepOver() {
        if( !inside )
        {
            LayoutState oldState = status;
            StageStatus stage = StageStatus.UNFINISHED;
            while( status == oldState && stage == StageStatus.UNFINISHED )
                stage = forwardStep();
            return stage;
        }
        else
            return forward( "forwardStepOver" );
    }

    @Override
    public StageStatus forwardStepOut() {
        if( !inside )
        {
            StageStatus status = StageStatus.UNFINISHED;
            while( status == StageStatus.UNFINISHED )
                status = forwardStep();
            return status;
        }
        else
            return forward( "forwardStepOut" );
    }

    @Override
    public StageStatus backwardStepOver() {
        if( !inside )
        {
            LayoutState oldState = status;
            StageStatus stage = StageStatus.UNFINISHED;
            while( status == oldState && stage == StageStatus.UNFINISHED )
                stage = backwardStep();
            return stage;
        }
        else
            return backward( "backwardStepOver" );
    }

    @Override
    public StageStatus backwardStepOut() {
        if( !inside )
        {
            StageStatus status = StageStatus.UNFINISHED;
            while( status == StageStatus.UNFINISHED )
                status = backwardStep();
            return status;
        }
        else
            return backward( "backwardStepOut" );
    }

    @Override
    public String getDebugString()
    {
        String info = "| Node | Shift | Sink | Root | Align |  x  |  xDef  |\n";
        info +=       "|------|-------|------|------|-------|-----|--------|\n";
        for( LayeredGraphNode n : graph.getContainedNodes() )
        {
            info += "|" + TextLayoutHelper.strToLen( n.getName(), 6 ) + 
                    "|" + TextLayoutHelper.strToLen( n.getShift( type ) + "", 7 ) + 
                    "|" + TextLayoutHelper.strToLen( n.getSink( type ).getName(), 6 ) + 
                    "|" + TextLayoutHelper.strToLen( n.getRoot( type ).getName(), 6 ) + 
                    "|" + TextLayoutHelper.strToLen( n.getAlign( type ).getName(), 7 ) + 
                    "|" + TextLayoutHelper.strToLen( n.getX( type ) + "", 5 ) + 
                    "|" + TextLayoutHelper.strToLen( !n.isXUndefined( type ) + "", 8 ) + "|\n";
        }
        if( inside && status == LayoutState.BLOCK_CALCULATION )
            info += bc.getDebugString();
        return info;
    }
    
    private StageStatus forward( String fName )
    {
        boolean breakpoint = false;
        CodeAction action = null;
        try {
            if( status == LayoutState.BLOCK_CALCULATION )
            {
                boolean selected = bcNode.isSelected();
                action = bcNode.setSelected( true );
                if( !selected )
                    breakpoint |= action == CodeAction.STOP;
                do {
                    switch( (StageStatus)(BlockCalc.class.getMethod( fName ).invoke( bc ) ) )
                    {
                    case FINISHED:
                        inside = false;
                        bcNode.setSelected( false );
                        CodeAction ac = cpNode.setSelected( true );
                        status = LayoutState.COMPACTION;
                        if( ac == CodeAction.SKIP && !breakpoint )
                            return forward( fName );
                        break;
                    case BREAKPOINT:
                        inside = true;
                        return StageStatus.BREAKPOINT;
                    case UNFINISHED:
                        inside = true;
                    }
                } while( !breakpoint && action == CodeAction.SKIP );
            }
            else if( status == LayoutState.COMPACTION )
            {
                boolean selected = cpNode.isSelected();
                action = cpNode.setSelected( true );
                if( !selected )
                    breakpoint |= action == CodeAction.STOP;
                do {
                    switch( (StageStatus)(Compaction.class.getMethod( fName ).invoke( cp ) ) )
                    {
                    case FINISHED:
                        inside = false;
                        cpNode.setSelected( false );
                        return StageStatus.FINISHED;
                    case BREAKPOINT:
                        inside = true;
                        return StageStatus.BREAKPOINT;
                    case UNFINISHED:
                        inside = true;
                    }
                } while( !breakpoint && action == CodeAction.SKIP );
            }
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
        } catch (SecurityException e) {
            e.printStackTrace();
        }
        if( breakpoint )
            return StageStatus.BREAKPOINT;
        return StageStatus.UNFINISHED;
    }
    
    private StageStatus backward( String fName )
    {
        boolean breakpoint = false;
        CodeAction action = null;
        try {
            if( status == LayoutState.BLOCK_CALCULATION )
            {
                boolean sel = bcNode.isSelected();
                action = bcNode.setSelected( true );
                if( !sel )
                    breakpoint |= action == CodeAction.STOP;
                do {
                    switch( (StageStatus)(BlockCalc.class.getMethod( fName ).invoke( bc ) ) )
                    {
                    case FINISHED:
                        inside = false;
                        bcNode.setSelected( false );
                        return StageStatus.FINISHED;
                    case BREAKPOINT:
                        inside = true;
                        return StageStatus.BREAKPOINT;
                    case UNFINISHED:
                        inside = true;
                    }
                } while( !breakpoint && action == CodeAction.SKIP );
            }
            else if( status == LayoutState.COMPACTION )
            {
                boolean sel = cpNode.isSelected();
                action = cpNode.setSelected( true );
                if( !sel )
                    breakpoint |= action == CodeAction.STOP;
                do {
                    switch( (StageStatus)(Compaction.class.getMethod( fName ).invoke( cp ) ) )
                    {
                    case FINISHED:
                        inside = false;
                        cpNode.setSelected( false );
                        status = LayoutState.BLOCK_CALCULATION;
                        action = null;
                        break;
                    case BREAKPOINT:
                        inside = true;
                        return StageStatus.BREAKPOINT;
                    case UNFINISHED:
                        inside = true;
                    }
                } while( !breakpoint && action == CodeAction.SKIP );
            }
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
        } catch (SecurityException e) {
            e.printStackTrace();
        }
        if( breakpoint )
            return StageStatus.BREAKPOINT;
        return StageStatus.UNFINISHED;
    }
}
